/*! \file  SEfile.h
 *  \brief This is the SEfile header that contains most of the methods, constants and structures used by SEfile.
 *  \author Fornero Matteo
 *  \date 25/11/2019
 */

#ifndef SEFILE_H_
#define SEFILE_H_

#include "../sources/L1/L1.h"
#include "SEfile_C_interface.h"

#define KEY_CHECK_INTERVAL 1 /**<  @brief Time interval (in seconds) used to check for the validity of the key used to encrypt the file. */
#define SEFILE_NONCE_LEN 32
extern bool override_key_check; /**<  @brief Global flag that is used to bypass the validity check of a key to read or write encrypted data. It is used only to re-encrypt data belonging to a compromised file. */

/**  @brief Length of header sector reserved to SEkey informations.
 * @details This is the length of the header required by SEkey. it is embedded in the SEfile header;
 * the length of the SEkey header must be a multiple of the block size used by the cipher
 * (16 bytes for AES256 at the moment) because the size of the data to cipher inside the
 * first sector is 512 - SEFILE_NONCE_LEN - SEKEY_HDR_LEN and it must be a multiple of 16. */
#define SEKEY_HDR_LEN 16

/** @brief The SEFILE_HANDLE struct
 * This abstract data type is used to hide from higher level of abstraction its implementation. The data stored in here are the current physical file pointer position and the file descriptor OS-dependent data type. */
#pragma pack(push,1)
struct SEFILE_HANDLE {
    uint32_t log_offset;    /**< Actual pointer position in bytes*/
#if defined(__linux__) || defined(__APPLE__)
    int32_t fd;             /**< File descriptor in Unix environment*/
#elif _WIN32
    HANDLE fd;              /**< File descriptor in Windows environment*/
#endif
    uint8_t nonce_ctr[16];  /**< Nonce used for the CTR feedback*/
    uint8_t nonce_pbkdf2[SEFILE_NONCE_LEN]; /**< Nonce used for the PBKDF2*/
    char name[MAX_PATHNAME]; /*< String that contains the name of the file. This is exploited only by the SEcure Database in order to run other databases apart from the one of SEkey. */
    SEFILE_HANDLE();
};

/** @brief The header of each file encrypted with SEfile.
 * This is the header that contains info for the SEkey system. It is stored as cleartext and it is not authenticated.
 * The dimension is always \ref SEKEY_HDR_LEN. */
struct SEKEY_HEADER {
	uint32_t key_id;		/**< The ID of the key used to encrypt this file. */
	uint16_t algorithm;		/**< The algorithm used to encrypt this file. */
	uint8_t padding[10];	/**< Padding used to reach a SEkey header size which is a multiple of 16. */
};

/** @brief The SEFILE_HEADER struct
 * This data struct is used to define a 31 bytes field inside a sector
 * while taking care of its inner composition. */
struct SEFILE_HEADER {
	uint8_t nonce_pbkdf2[SEFILE_NONCE_LEN];	/**< 32 random bytes storing the IV for generating a different key*/
	SEKEY_HEADER key_header; /**< The header with the ID of the key and the algorithm. */
    uint8_t nonce_ctr[16];		            /**< 16 random bytes storing the IV for next sectors*/
    int32_t magic;				            /**< 4 bytes used to represent file type (not used yet)*/
    int16_t ver;				            /**< 2 bytes used to represent current filesystem version (not used yet)*/
    int32_t uid;				            /**< 4 bytes not used yet*/
    int32_t uid_cnt;			            /**< 4 bytes not used yet*/
    uint8_t fname_len;			            /**< 1 byte to express how long is the filename.*/
};

/** @brief The SEFILE_SECTOR struct
 * This data struct is the actual sector organization.
 * The total size should ALWAYS be equal to \ref SEFILE_SECTOR_SIZE.
 * The first sector is used to hold ONLY the header.
 * Thanks to the union data type, the developer can simply declare
 * a sector and then choose if it is the header sector or not. */
struct SEFILE_SECTOR {
	union {
        SEFILE_HEADER header;               /**< See \ref SEFILE_HEADER .*/
        uint8_t data[SEFILE_LOGIC_DATA];   /**< In here it will be written the actual data.
         *   Since it is inside a union data type, the
         *   filename will be written from 32nd byte.
         */
    };
    uint16_t len;                      /**< How many bytes are actually stored in this sector.*/
    uint8_t signature[32];             /**< Authenticated digest generated by the device*/
    SEFILE_SECTOR(); /**< Constructor used to initialize all the fields of the struct to zero. */
};
#pragma pack(pop)

/* functions not related to SEfile objects that can be called by higher levels */
uint16_t get_secure_context(std::string& filename, std::string *keyid, uint16_t *algo);
uint16_t secure_mkdir(std::string& path, L1 *SEcubeptr, uint32_t key);
uint16_t secure_ls(std::string& path, std::vector<std::pair<std::string, std::string>>& list, L1 *SEcubeptr);
uint16_t secure_getfilesize(char *path, uint32_t * position, L1 *SEcubeptr);
uint16_t secure_recrypt(std::string path, uint32_t key, L1 *SEcubeptr);

/* functions that are not related to SEfile objects that should not be called by higher levels because they are used internally by SEfile methods */
uint16_t decrypt_filename(std::string& path, char *filename, L1 *SEcubeptr);
/** @brief This function is used to compute the ciphertext of a directory name stored in dirname.
* @param [in] path Path to the directory whose name has to be encrypted.
* @param [out] encDirname A preallocated string where to store the encrypted directory name.
* @param [out] enc_len Where to store how many bytes the encrypted directory name is long. Can be NULL.
* @param [in] SEcubeptr Pointer to the L1 object used to communicate with the SEcube.
* @param [in] key ID of the key to be used to encrypt the name of the new directory.
* @return The function returns 0 in case of success. See \ref errorValues for error list. */
uint16_t crypt_dirname(std::string& path, char *encDirname, uint32_t* enc_len, L1 *SEcubeptr, uint32_t key);
uint16_t decrypt_dirname(std::string& path, char *decDirname, L1 *SEcubeptr);
uint16_t encrypt_name(void* buff1, void* buff2, size_t size, uint16_t direction, L1 *SEcubeptr, uint32_t key); /**< @brief Internally used by crypt_dirname and decrypt_dirname(). */
uint16_t valid_directory_name(std::string& name);
uint16_t valid_file_name(std::string& name);
size_t pos_to_cipher_block(size_t current_position); /**< @brief Internally used by SEfile functions. */
void compute_blk_offset(size_t current_offset, uint8_t* nonce); /**< @brief Internally used by SEfile functions. */
void get_filename(char *path, char *file_name); /**< @brief Extract the name of a file from its path. */
void get_path(char *full_path, char *path); /**< @brief Extract the path of a file removing the file name. */

/**
* \class SEfile
* @brief A SEfile object is used to manage a file encrypted with SEfile.
* @details Each SEfile object has several attributes that define the security context. A security context is made of a key, an algorithm and a SEcube.
* These three parameters define how the file must be protected (i.e. AES-256 encrypted using the key with ID 10). There are also other attributes
* that are used for internal purposes. The class internally uses smart pointers and the RAII approach to guarantee correct memory management. Notice that
* the attributes are public just for ease of use.
*/
class SEfile{
public:
	 uint32_t EnvKeyID; /**< @brief The key ID used by this SEfile instance. This key will be used for encryption and decryption. */
	 uint16_t EnvCrypto; /**<  @brief The algorithm to be used with the key. */
	 time_t LastEncryptCheckTime; /**<  @brief The last time the validity of the key used by this file was checked, requiring write (encrypt, requires active key) privilege. */
	 time_t LastDecryptCheckTime; /**<  @brief The last time the validity of the key used by this file was checked, requiring read (decrypt, does not require active key) privilege. */
	 bool IsOpen; /**<  @brief Flag that is TRUE if the file is open, FALSE otherwise. */
	 L1 *l1; /**<  @brief The pointer to the L1 object created by the main application (i.e. to login to the SEcube). */
	 std::shared_ptr<SEFILE_HANDLE> handleptr; /**<  @brief Handle to the file on the underlying filesystem. */
	 /* Notice that a shared_ptr is used for the SEFILE_HANDLE structure because it is more manageable by other components of the SEcube SDK (i.e. SEkey KMS and the SEcure Database).
	  * Considering SEfile only, having a pointer or having directly the structure inside the class makes no difference...but it makes a difference when using SEfile together with SQLite
	  * for the SEcure Database. Therefore, in order to keep the same object for the normal SEfile version and for the SEfile of the SEcure DB, the smart pointer is better. */
	 SEfile(); /**<  @brief Default constructor. Initializes the secure environment with empty values. */
	 SEfile(L1 *secube); /**<  @brief Constructor to initialize the secure environment with empty values, apart from the pointer to the SEcube to be used. */
	 SEfile(L1 *secube, uint32_t keyID); /**<  @brief Constructor to initialize the secure environment with empty values, apart from the pointer to the SEcube to be used and the ID of the key to be used. */
	 SEfile(L1 *secube, uint32_t keyID, uint16_t crypto); /**<  @brief Constructor to fully initialize the secure environment. */
	 ~SEfile(); /**<  @brief Destructor. Automatically calls secure_finit() and secure_close(). */
	 /* APIs of SEfile */
	 uint16_t secure_init(L1 *l1ptr, uint32_t keyID, uint16_t crypto);
	 void secure_finit();
	 /**
	  * @defgroup sefile_apis
	  *  @{
	  */
	 	/**
	 	 * \name APIs of SEfile
	 	 * \details These APIs are the most useful to exploit the SEfile library.
	 	 */
	 	///@{
			 uint16_t secure_open(char *path, int32_t mode, int32_t creation);
			 uint16_t secure_close();
			 uint16_t secure_read(uint8_t *dataOut, uint32_t dataOut_len, uint32_t *bytesRead);
			 uint16_t secure_write(uint8_t *dataIn, uint32_t dataIn_len);
			 uint16_t secure_seek(int32_t offset, int32_t *position, uint8_t whence);
			 uint16_t secure_truncate(uint32_t size);
			 uint16_t secure_sync();
		///@}
	/** @}*/
	 /**
	  * @defgroup sefile_internal_functions
	  *  @{
	  */
	 	/**
	 	 * \name Functions for internal purposes of SEfile
	 	 * \details You should not need to use these functions directly because they are used internally by the SEfile library.
	 	 */
	 	///@{
			 uint16_t secure_create(char *path, std::shared_ptr<SEFILE_HANDLE> hFile, int mode);
			 uint16_t secure_key_check(uint16_t direction);
			 uint16_t get_filesize(uint32_t * length);
			 uint16_t crypt_header(void *buff1, void *buff2, size_t datain_len, uint16_t direction);
			 uint16_t crypt_sectors(void *buff_decrypt, void *buff_crypt, size_t datain_len, size_t current_offset, uint8_t* nonce_ctr, uint8_t* nonce_pbkdf2);
			 uint16_t decrypt_sectors(void *buff_crypt, void *buff_decrypt, size_t datain_len, size_t current_offset, uint8_t* nonce_ctr, uint8_t* nonce_pbkdf2);
		///@}
	/** @}*/
	 /**
	  * @defgroup sefile_custom_sqlite_apis
	  *  @{
	  */
	 	/**
	 	 * \name APIs of SEfile for SQLite DB engine
	 	 * \details These APIs customized to apply SEfile to the SQLite db engine. They are called automatically by the custom VFS of SQLite;
	      * they are required by SEkey and by the SEcure Database library. You should not need to use these APIs directly.
	 	 */
	 	///@{
			 uint16_t securedb_secure_open(char *path, int32_t mode, int32_t creation);
			 uint16_t securedb_secure_close();
			 uint16_t securedb_secure_read(uint8_t * dataOut, uint32_t dataOut_len, uint32_t *bytesRead);
			 uint16_t securedb_secure_write(uint8_t * dataIn, uint32_t dataIn_len);
			 uint16_t securedb_secure_seek(int32_t offset, int32_t *position, uint8_t whence);
			 uint16_t securedb_secure_truncate(uint32_t size);
			 uint16_t securedb_secure_sync();
			 uint16_t securedb_get_filesize(uint32_t * length);
			 uint16_t securedb_secure_create(char *path, std::shared_ptr<SEFILE_HANDLE> hFile, int mode);
		///@}
	/** @}*/
};

#endif
